/**
 * libFAD - Flash Animation Decode library
 * Copyright (C) 2005-2006 VGSystem Technologies, Inc.
 *
 * libFAD is the legal property of its developers, whose names are too numerous
 * to list here.  Please refer to the COPYRIGHT file distributed with this
 * source distribution.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: fad.shape.c,v 1.72 2006/05/08 06:58:18 wrxzzj Exp $
 */
#include "fad.shape.h"
#include "fad.tags.h"
#include "fad.bits.h"

#define FAD_ERROR

typedef struct {
  fad_object_t base;
  rect_t bound;
  shape_record_t sr;
} style_shape_t;

static u8_t _style_shape_do_render(fad_object_t* fo, fad_render_t* render, dl_node_t *node);
static void _style_shape_do_free(fad_object_t* fo);
static void  _style_shape_get_rect(fad_object_t* fo, rect_t *rt);

static cairo_pattern_t *pat = NULL;
static cairo_surface_t *ips = NULL;

static void _line_style_do_render(fad_render_t* render, line_style_array_t* lsa, u16_t idx, u8_t srv, dl_node_t *node) {
  u8_t* ptr = NULL, r, g, b, a;
  u16_t width;

  if(lsa && idx < lsa->count) {
    ptr = lsa->array;

    if(srv == FAD_SRV_SHAPE3) {
      ptr += idx*6;
      r = *(ptr+2); g = *(ptr+3); b = *(ptr+4); a = *(ptr+5);
      if(node->flag&0x08) {
        r = COLOR_DO_CXFORM(r, node->cx.base.rm, node->cx.base.ra);
        g = COLOR_DO_CXFORM(g, node->cx.base.gm, node->cx.base.ga);
        b = COLOR_DO_CXFORM(b, node->cx.base.bm, node->cx.base.ba);
        a = COLOR_DO_CXFORM(a, node->cx.am, node->cx.aa);
      }

      width = *ptr|*(ptr+1)<<8;
      if(width < 1)
        width = 1;
      cairo_set_line_width(render->cr, (double)width/20);
    } else {
      ptr += idx*5;
      r = *(ptr+2); g = *(ptr+3); b = *(ptr+4); a = 0xff;
      if(node->flag&0x08) {
        r = COLOR_DO_CXFORM(r, node->cx.base.rm, node->cx.base.ra);
        g = COLOR_DO_CXFORM(g, node->cx.base.gm, node->cx.base.ga);
        b = COLOR_DO_CXFORM(b, node->cx.base.bm, node->cx.base.ba);
        a = COLOR_DO_CXFORM(a, node->cx.am, node->cx.aa);
      }
      
      width = *ptr|*(ptr+1)<<8;
      if(width < 2)
        width = 2;
      cairo_set_line_width(render->cr, (double)width/20);
    }

    FAD_ERROR("style-shape set line color(r = %d, g = %d, b = %d, a = %d), line width = %d.\n", r, g, b, a, width);
    cairo_set_source_rgba(render->cr, (double)r/0xff, (double)g/0xff, (double)b/0xff, (double)a/0xff);
  }
}

static void _fill_style_do_render(fad_render_t* render, fill_style_array_t* fsa, u16_t idx, u8_t srv, dynarray_t* dict, dl_node_t *node) {
  u8_t* ptr = NULL, type, r, g, b, a;
  bits_t bits;
  cairo_matrix_t mx;

  if(fsa && idx < fsa->count) {
    FAD_ERROR("fill style do render (%d, fsa = %x)\n", idx, fsa);
    bits_init(&bits);
    ptr = fsa->array[idx];

    switch(type = *ptr++) {
      case 0x00:
        r = *ptr; g = *(ptr+1); b = *(ptr+2); a = 0xff;

        if(srv == FAD_SRV_SHAPE3)
          a = *(ptr+3);

        if(node->flag&0x08) {
          r = COLOR_DO_CXFORM(r, node->cx.base.rm, node->cx.base.ra);
          g = COLOR_DO_CXFORM(g, node->cx.base.gm, node->cx.base.ga);
          b = COLOR_DO_CXFORM(b, node->cx.base.bm, node->cx.base.ba);
          a = COLOR_DO_CXFORM(a, node->cx.am, node->cx.aa);
        }
        FAD_ERROR("style-shape set fill color (%d, %d, %d, %d), fsa->count = %d.\n", r, g, b, a, fsa->count);

        cairo_set_source_rgba(render->cr, (double)r/0xff, (double)g/0xff, (double)b/0xff, (double)a/0xff);
        break;

      case 0x10:
      case 0x12:
        {
          u8_t num;

          if(pat != NULL) {
            cairo_pattern_destroy(pat);
            pat = NULL;
          }

          bits_buffer(&bits, ptr);
          bits_get_matrix(&bits, &mx);
          if(type == 0x10) {
            pat = cairo_pattern_create_linear(-16384.00/20.00, 0.00, 16384.00/20.00, 0.00);
            FAD_ERROR("style-shape set fill linear gradient.\n");
          } else {
            pat = cairo_pattern_create_radial(0.00, 0.00, 0.00, 0.00, 0.00, 16384.00/20.00);
            FAD_ERROR("style-shape set fill radial gradient.\n");
          }

          FAD_ERROR("gradient pattern matrix = (%f, %f, %f, %f, %f, %f)\n", mx.xx, mx.xy, mx.x0, mx.yy, mx.yx, mx.y0);
          ptr = bits_tell(&bits);
          for(num = *ptr++; num > 0; num--) {
            if(srv == FAD_SRV_SHAPE3) {
              FAD_ERROR("add color stop RGBA(%x, %x, %x, %x)\n", *ptr, *(ptr+1), *(ptr+2), *(ptr+3), *(ptr+4));
              cairo_pattern_add_color_stop_rgba(pat, (double)*ptr/0xff, 
                  (double)*(ptr+1)/0xff, (double)*(ptr+2)/0xff, (double)*(ptr+3)/0xff, (double)*(ptr+4)/0xff);
              ptr += 5;
            } else {
              FAD_ERROR("add color stop RGB(%x, %x, %x, 0xff)\n", *ptr, *(ptr+1), *(ptr+2), *(ptr+3));
              cairo_pattern_add_color_stop_rgb(pat, (double)*ptr/0xff,
                  (double)*(ptr+1)/0xff, (double)*(ptr+2)/0xff, (double)*(ptr+3)/0xff);
              ptr += 4;
            }
          }
          cairo_matrix_invert(&mx);
          cairo_pattern_set_matrix(pat, &mx);
          cairo_set_source(render->cr, pat);
        }
        break;

      case 0x40:
      case 0x41:
      case 0x42:
      case 0x43:
        {
          fad_object_t* fo = NULL;

          bits_buffer(&bits, ptr);
          fo = dict->get(dict, bits_get_u16(&bits));

#if 1
          if(fo && fo->get_image) {
            s32_t w, h;
            cairo_matrix_t mx;
            void* image = fo->get_image(fo, &w, &h);

            //FAD_ERROR("style shape set fill bitmap, fo = %x, image = %x\n", fo, image);
            if(pat != NULL)
              cairo_pattern_destroy(pat);
            if(ips != NULL)
              cairo_surface_destroy(ips);

            if(image) {
              bits_get_matrix(&bits, &mx);
              mx.xx = mx.xx/20.00;
              mx.yy = mx.yy/20.00;
              FAD_ERROR("image pattern matrix = (%f, %f, %f, %f, %f, %f), type = %x\n", mx.xx, mx.xy, mx.x0, mx.yy, mx.yx, mx.y0, type);
              ips = cairo_image_surface_create_for_data(image, CAIRO_FORMAT_ARGB32, w, h, w<<2);
              pat = cairo_pattern_create_for_surface(ips);
              if(type == 0x40 || type == 0x42)
                cairo_pattern_set_extend(pat, CAIRO_EXTEND_REPEAT);
              else
                cairo_pattern_set_extend(pat, CAIRO_EXTEND_NONE);

              cairo_matrix_invert(&mx);
              cairo_pattern_set_matrix(pat, &mx);
              cairo_set_source(render->cr, pat);
            }
          }
#endif
        }
        break;

      default:
        FAD_ERROR("style-shape fill style unknown, type = %x\n", type);
        break;
    }
    bits_finish(&bits);
  }
}

static u16_t _line_style_array_decode(line_style_array_t* lsa, bits_t* bits, u8_t srv) {
  u32_t offset = 0;

  lsa->count = bits_get_u8(bits);
  lsa->array = NULL;
  if(lsa->count == 0xFF)
    lsa->count = bits_get_u16(bits);

  lsa->array = (u8_t* )bits_tell(bits);
  FAD_ERROR("line style array count = %d, array = %x\n", lsa->count, lsa->array);
  if(lsa->count > 0) {
    if(srv == FAD_SRV_SHAPE3)
      offset = 6;
    else
      offset = 5;

    offset = offset * lsa->count;
    bits_seek_nbytes(bits, offset);
  }

  return lsa->count;
}

static u16_t _fill_style_array_decode(fill_style_array_t* fsa, bits_t* bits, u8_t srv) {
  u16_t idx = 0, offset = 0;

  fsa->count = bits_get_u8(bits);
  fsa->array = NULL;
  if(fsa->count == 0xFF)
    fsa->count = bits_get_u16(bits);

  if(fsa->count > 0) {
    fsa->array = calloc(fsa->count, sizeof(u8_t* ));
    FAD_ERROR("fill style array count = %d, array = %x\n", fsa->count, fsa->array);

    do {
      fsa->array[idx] = (u8_t* )bits_tell(bits);
      //FAD_ERROR("fill style array[%d] = %x\n", idx, bits_tell(bits));
      switch(bits_get_u8(bits)) {
        case 0x00:
          if(srv == FAD_SRV_SHAPE3)
            offset = 4;
          else
            offset = 3;
          bits_seek_nbytes(bits, offset);
          break;

        case 0x10:
        case 0x12:
          bits_get_matrix(bits, NULL);
          if(srv == FAD_SRV_SHAPE3)
            offset = 5;
          else
            offset = 4;

          offset = offset*bits_get_u8(bits);
          bits_seek_nbytes(bits, offset);
          break;

        case 0x40:
        case 0x41:
        case 0x42:
        case 0x43:
          bits_seek_nbytes(bits, 2);
          bits_get_matrix(bits, NULL);
          break;
      }
    } while(++idx < fsa->count);
  }

  return fsa->count;
}

s32_t style_shape_decode(fad_frame_t* frame, fad_stream_t* s) {
  style_shape_t* ss = NULL;
  u16_t id;

  ss = calloc(1, sizeof(style_shape_t));
  if(ss == NULL) {
    s->err = FAD_ERROR_MEM;
    return -1;
  }

  ss->base.do_render = _style_shape_do_render;
  ss->base.do_free = _style_shape_do_free;
  ss->base.get_rect= _style_shape_get_rect;
  ss->base.type = FO_TYPE_STYLESHAPE;

  id = bits_get_u16(&s->bits);

  bits_get_rect(&s->bits, &ss->bound);
  ss->sr.dict = s->dict;

  if(s->tag_id == TAG_DEFINESHAPE)
    ss->sr.srv = FAD_SRV_SHAPE1;
  else if(s->tag_id == TAG_DEFINESHAPE2)
    ss->sr.srv = FAD_SRV_SHAPE2;
  else if(s->tag_id == TAG_DEFINESHAPE3)
    ss->sr.srv = FAD_SRV_SHAPE3;

  shape_record_parse(&s->bits, &ss->sr);
  s->dict->put(s->dict, ss, id);
  FAD_ERROR("append style-shape = %x to dictionary, id = %d, tag_id = %d\n", ss, id, s->tag_id);

  return 0;
}

static sr_node_t* sr_node_new(s32_t x0, s32_t y0, s32_t x1, s32_t y1, u8_t type) {
  sr_node_t *node = NULL;

  node = calloc(1, sizeof(sr_node_t));
  node->type = type;
  node->x0 = x0; node->y0 = y0;
  node->x1 = x1; node->y1 = y1;
  node->next = NULL;

  return node;
}

static void sr_node_destroy(sr_node_t* node) {
  if(node != NULL)
    free(node);
}

static sr_node_list_t *sr_node_list_new(sr_node_t *node) {
  sr_node_list_t *nl = NULL;

  nl = calloc(1, sizeof(sr_node_list_t));
  nl->header = node;
  nl->tail = node;
  nl->sta  = FAD_NLSTA_INIT;

  return nl;
}

static void sr_node_list_destroy(sr_node_list_t *nl) {
  sr_node_t *ptr = NULL, *tmp = NULL;

  for(ptr = nl->header; ptr != NULL; ) {
    tmp = ptr;
    ptr = ptr->next;
    sr_node_destroy(tmp);
    tmp = NULL;
  }
}

static void sr_node_list_rebuild(sr_node_list_t *nl) {
  sr_node_t *ptr1 = NULL, *ptr2;

  if(nl != NULL) {
    ptr1 = nl->header;
    ptr2 = ptr1->next;

    if(ptr1->type == FAD_SRT_CPT && ptr2->type == FAD_SRT_LPT) {

      ptr2->x1 = ptr1->x1; ptr2->y1 = ptr1->y1;
      ptr1->x1 = 0;        ptr1->y1 = 0;

      ptr1->type = FAD_SRT_LPT;
      ptr2->type = FAD_SRT_CPT;
    }
  }
}

static void sr_node_list_append(sr_node_list_t *nl, sr_node_t *node) {
  if(nl != NULL) {
    if(nl->header != NULL) {
      nl->tail->next = node;
      nl->tail = node;
    } else {
      nl->header = node;
      nl->tail = node;
    }
  }
}

static void sr_node_list_prepend(sr_node_list_t *nl, sr_node_t *node) {
  if(nl != NULL) {
    if(nl->header != NULL) {
      node->next = nl->header;
      nl->header = node;
    } else {
      nl->header = node;
      nl->tail = node;
    }
  }
}

static void sr_node_list_merge(sr_edgesrecord_t *er, sr_node_list_t *nl, u16_t idx) {
  sr_node_list_t *ptr = er->array[idx], *tail = NULL;
  s8_t match = FAD_FALSE;

  DEBUG_ENTER;
  if(nl->sta == FAD_NLSTA_INIT) {
    if(ptr == NULL) {
      nl->next = nl->prev = nl;
      ptr = er->array[idx] = nl;
    } else {
      tail = ptr->prev;

      tail->next = nl;
      ptr->prev = nl;
      nl->prev = tail;
      nl->next = ptr;
    }
    nl->sta = FAD_NLSTA_OPEN;
  }

  for(ptr=nl->next; ptr!=nl; ptr=ptr->next) {
    if(ptr->sta == FAD_NLSTA_OPEN) {
      if(ptr->header->x0==nl->tail->x0 && ptr->header->y0==nl->tail->y0) {
        nl->tail->next = ptr->header->next;
        free(ptr->header);
        nl->tail = ptr->tail;

        ptr->prev->next = ptr->next;
        ptr->next->prev = ptr->prev;
        if(ptr == er->array[idx]) {
          er->array[idx] = ptr->next;
        }
        free(ptr);
        ptr = nl;
      } else if(ptr->tail->x0==nl->header->x0 && ptr->tail->y0==nl->header->y0) {
        ptr->tail->next = nl->header->next;
        free(nl->header);
        ptr->tail = nl->tail;

        nl->prev->next = nl->next;
        nl->next->prev = nl->prev;
        if(nl == er->array[idx]) {
          er->array[idx] = nl->next;
        }
        free(nl);
        nl = ptr;
      }
    }
  }

  if((nl->sta==FAD_NLSTA_OPEN) && (nl->header!=nl->tail)) {
    if(nl->header->x0==nl->tail->x0 && nl->header->y0==nl->tail->y0)
      nl->sta = FAD_NLSTA_CLOSE;
  }
  DEBUG_LEVEL;
}

static sr_node_list_t *sr_node_list_search_by_point(sr_node_list_t *er, s32_t x, s32_t y, u8_t dir) {
  sr_node_t *node = NULL;
  u8_t match = FAD_FALSE;
  sr_node_list_t *ptr = er;

  if(ptr == NULL)
    return NULL;

  DEBUG_ENTER;
  FAD_ERROR("search edgesrecord hdr = %x\n", er);
  do {
    if(ptr->sta == FAD_NLSTA_OPEN) {
      if(dir == FAD_DIRT_FALSE) {
        if(ptr->header->x0==x && ptr->header->y0==y) {
          match = FAD_TRUE;
          break;
        }
      } else {
        if(ptr->tail->x0==x && ptr->tail->y0==y) {
          match = FAD_TRUE;
          break;
        }
      }
    }
    ptr = ptr->next;
  } while(ptr != er);

  DEBUG_LEVEL;

  return match==FAD_TRUE ? ptr : NULL;
}

static sr_edgesrecord_t * sr_edgesrecord_new(u16_t fs_num, u16_t ls_num, sr_newstyle_t *ns) {
  sr_edgesrecord_t *er = NULL;

  er = calloc(1, sizeof(sr_edgesrecord_t)+sizeof(sr_node_list_t*)*(fs_num+1+ls_num+1));
  er->fs_num = fs_num;
  er->ls_num = ls_num;
  er->ns = ns;

  return er;
}

static void sr_edgesrecord_set_ns(sr_edgesrecord_t *er, sr_newstyle_t *ns) {
  if(er != NULL)
    er->ns = ns;
  //FAD_ERROR("er = %x, ns = %x, er->ns = %x\n", er, ns, er->ns);
}

static void sr_edgesrecord_destroy(sr_edgesrecord_t *er) {
  u16_t idx;

  for(idx = 0; idx<er->fs_num+er->ls_num; idx++) {
    if(er->array[idx]) {
      sr_node_list_destroy(er->array[idx]);
      free(er->array[idx]);
      er->array[idx] = NULL;
    }
  }
}

static void edgesrecord_add_node_list(sr_edgesrecord_t *er, u16_t idx, sr_node_list_t *nl) {
  sr_node_list_t *ptr = NULL, *tail= NULL;

  ptr = er->array[idx];

  /**edges record node list is a circle list*/
  if(ptr != NULL) {
    tail = ptr->prev;

    tail->next = nl;
    ptr->prev = nl;
    nl->prev = tail;
    nl->next = ptr;
  } else {
    nl->next = nl;
    nl->prev = nl;
    er->array[idx] = nl;
  }
}

s32_t shape_record_parse(bits_t* bits, shape_record_t* sr) {
  sr_newstyle_t *ns = NULL;
  u8_t flag, nbits, fs_nbits, ls_nbits, type = FAD_SRT_XXX;
  u16_t fs0_idx, fs1_idx, ls_idx, fs0 = 0, fs1 = 0, ls = 0;
  s32_t x = 0, y = 0;
  sr_edgesrecord_t *er = NULL;
  u16_t fs_num = 1, ls_num = 0;
  sr_node_list_t *fs0_nl = NULL, *fs1_nl = NULL, *ls_nl = NULL;
  static s32_t node_count = 0;

  if(sr->srv == FAD_SRV_SHAPE1 || sr->srv == FAD_SRV_SHAPE2 || sr->srv == FAD_SRV_SHAPE3) {
    ns = calloc(1, sizeof(sr_newstyle_t));
    if(ns == NULL)
      return SHAPE_RECORD_PARSE_FALSE;

    fs_num = _fill_style_array_decode(&ns->fs_array, bits, sr->srv);
    ls_num = _line_style_array_decode(&ns->ls_array, bits, sr->srv);
  }

  er = sr_edgesrecord_new(fs_num, ls_num, ns);
  if(sr->header == NULL) {
    sr->header = sr->tail = er;
  }

  fs_nbits = bits_get_u8(bits);
  ls_nbits = fs_nbits&0x0f;
  fs_nbits = fs_nbits>>4;
  //FAD_ERROR("fs_nbits = %d, ls_nbits = %d\n", fs_nbits, ls_nbits);

  while(flag = bits_get_ubits(bits, 6)) {
    if(flag&0x20) {
      nbits = (flag&0x0f)+2;
      if(flag&0x10) {
        if(bits_get_ubits(bits, 1)) {
          x += bits_get_sbits(bits, nbits);
          y += bits_get_sbits(bits, nbits);
        } else {
          if(bits_get_ubits(bits, 1))
            y += bits_get_sbits(bits, nbits);
          else
            x += bits_get_sbits(bits, nbits);
        }
        type = FAD_SRT_LPT;

        if(fs1 != fs0) {
          if(fs0_nl) {
            sr_node_list_append(fs0_nl, sr_node_new(x, y, 0, 0, type));
          } else if(fs0 > 0) {
            fs0_nl = sr_node_list_new(sr_node_new(x, y, 0, 0, type));
          }
          if(fs1_nl) {
            sr_node_list_prepend(fs1_nl, sr_node_new(x, y, 0, 0, type));
          } else if(fs1 > 0) {
            fs1_nl = sr_node_list_new(sr_node_new(x, y, 0, 0, type));
          }
        }

        if(ls_nl) {
          sr_node_list_append(ls_nl, sr_node_new(x, y, 0, 0, type));
        } else if(ls>0 && ls<=ls_num) {
          ls_nl = sr_node_list_new(sr_node_new(x, y, 0, 0, type));
        }

        FAD_ERROR("append line to(%f, %f) to shape record, fs0_nl=%x, fs1_nl = %x, ls_nl = %x, node.count = %d.\n", x/20.00, y/20.00, fs0_nl, fs1_nl, ls_nl, node_count++);
      } else {
        s32_t ctlx, ctly;

        x += bits_get_sbits(bits, nbits);
        y += bits_get_sbits(bits, nbits);
        ctlx = x;
        ctly = y;
        x += bits_get_sbits(bits, nbits);
        y += bits_get_sbits(bits, nbits);
        type = FAD_SRT_CPT;

        if(fs1 != fs0) {
          if(fs0_nl) {
            sr_node_list_append(fs0_nl, sr_node_new(x, y, ctlx, ctly, type));
          } else if(fs0 > 0) {
            fs0_nl = sr_node_list_new(sr_node_new(x, y, ctlx, ctly, type));
          }
          if(fs1_nl) {
            sr_node_list_prepend(fs1_nl, sr_node_new(x, y, ctlx, ctly, type));
            sr_node_list_rebuild(fs1_nl);
          } else if(fs1 > 0) {
            fs1_nl = sr_node_list_new(sr_node_new(x, y, ctlx, ctly, type));
          }
        }
        if(ls_nl) {
          sr_node_list_append(ls_nl, sr_node_new(x, y, ctlx, ctly, type));
        } else if(ls>0 && ls<=ls_num) {
          ls_nl = sr_node_list_new(sr_node_new(x, y, ctlx, ctly, type));
        }

        FAD_ERROR("append curve to(%f, %f, %f, %f) to shape record, fs0_nl = %x, fs1_nl = %x, ls_nl = %x, node.count=%d.\n", ctlx/20.00, ctly/20.00, x/20.00, y/20.00, fs0_nl, fs1_nl, ls_nl, node_count++);
      }
    } else {
      u8_t fs_change = FAD_FALSE;

      if(fs0_nl && ((flag&0x01)||(flag&0x02))) {
        sr_node_list_merge(er, fs0_nl, fs0);
        fs0_nl = NULL;
      }
      if(fs1_nl && ((flag&0x01)||(flag&0x04))) {
        sr_node_list_merge(er, fs1_nl, fs1);
        fs1_nl = NULL;
      }
      if(ls_nl && ((flag&0x01)||(flag&0x08))) {
        sr_node_list_merge(er, ls_nl, fs_num+ls);
        ls_nl = NULL;
      }

      if(flag&0x01) {
        nbits = bits_get_ubits(bits, 5);
        x = bits_get_sbits(bits, nbits);
        y = bits_get_sbits(bits, nbits);
        type = FAD_SRT_LPT;
      }

      /**just hack the flag value, when fs0 == fs1*/
      if(fs0 == fs1)
        fs_change = FAD_TRUE;

      if(flag&0x02) {
        fs0 = bits_get_ubits(bits, fs_nbits);
        if(fs0 <= 0) {
          fs0_nl = NULL;
        }
      }
      if(flag&0x04) {
        fs1 = bits_get_ubits(bits, fs_nbits);
        if(fs1 <= 0) {
          fs1_nl = NULL;
        }
      }
      if(flag&0x08) {
        ls = bits_get_ubits(bits, ls_nbits);
        if(ls <= 0) {
          ls_nl = NULL;
        }
      }

      if(fs0 != fs1) {
        if(fs0 > 0) {
          if(fs_change || (flag&0x01) || ((flag&0x02)&&(type!=FAD_SRT_XXX))) {
            fs0_nl = sr_node_list_search_by_point(er->array[fs0], x, y, FAD_DIRT_TRUE);
            if(fs0_nl == NULL)
              fs0_nl = sr_node_list_new(sr_node_new(x, y, 0, 0, FAD_SRT_LPT));
          }
        }
        if(fs1 > 0) {
          if(fs_change || (flag&0x01) || ((flag&0x04)&&(type!=FAD_SRT_XXX))) {
            fs1_nl = sr_node_list_search_by_point(er->array[fs1], x, y, FAD_DIRT_FALSE);
            if(fs1_nl==NULL)
              fs1_nl = sr_node_list_new(sr_node_new(x, y, 0, 0, FAD_SRT_LPT));
          }
        }
      } else {
        if(fs0_nl) {
          sr_node_list_merge(er, fs0_nl, fs0);
          fs0_nl = NULL;
        }
        if(fs1_nl) {
          sr_node_list_merge(er, fs1_nl, fs1);
          fs1_nl = NULL;
        }
      }

      if(ls>0 && ls<=ls_num) {
        if((flag&0x01) || ((flag&0x08)&&(type!=FAD_SRT_XXX))){
          ls_nl = sr_node_list_search_by_point(er->array[fs_num+ls], x, y, FAD_DIRT_TRUE);
          if(ls_nl==NULL)
            ls_nl = sr_node_list_new(sr_node_new(x, y, 0, 0, FAD_SRT_LPT));
        }
      }

      if(flag&0x01)
        FAD_ERROR("append move to(%f, %f) to shape record, fs0_nl = %x, fs1_nl = %x, ls_nl = %x, node_count = %d.\n", x/20.00, y/20.00, fs0_nl, fs1_nl, ls_nl, node_count++);

      if(flag&0x10) {
        ns = calloc(1, sizeof(sr_newstyle_t));
        fs_num = _fill_style_array_decode(&ns->fs_array, bits, sr->srv);
        ls_num = _line_style_array_decode(&ns->ls_array, bits, sr->srv);
        fs0 = fs1 = ls = 0;
        fs0_nl = fs1_nl = ls_nl = NULL;
        type = FAD_SRT_XXX;

        er = sr_edgesrecord_new(fs_num, ls_num, ns);

        if(sr->tail == NULL)
          sr->header = er;
        else
          sr->tail->next = er;
        sr->tail = er;

        fs_nbits = bits_get_u8(bits);
        ls_nbits = fs_nbits&0x0f;
        fs_nbits = fs_nbits>>4;
        FAD_ERROR("append new style to shape record, er = %x, ns = %x, fs_num = %d, ls_num = %d\n", er, ns, fs_num, ls_num);
      }
    }
  }

  if(fs0_nl) {
    sr_node_list_merge(er, fs0_nl, fs0);
  }
  if(fs1_nl) {
    sr_node_list_merge(er, fs1_nl, fs1);
  }
  if(ls_nl) {
    sr_node_list_merge(er, ls_nl, fs_num+ls);
  }

  return SHAPE_RECORD_PARSE_TRUE;
}

static void shape_record_render_node_list(sr_node_list_t *nl, fad_render_t *render, dl_node_t *node) {
  sr_node_list_t *ptr = nl;
  sr_node_t *hdr = NULL;
  s32_t x, y;

  FAD_ERROR("render node list = %x\n", ptr);
  hdr = ptr->header;
#ifdef LIBFAD_DO_RENDER
  cairo_move_to(render->cr, hdr->x0/20.00, hdr->y0/20.00);
#endif
  x = hdr->x0; y = hdr->y0;
  FAD_ERROR("move to (%f, %f)\n", hdr->x0/20.00, hdr->y0/20.00);
  hdr = hdr->next;
  
  while (hdr != NULL) {
    if(hdr->type == FAD_SRT_LPT) {
#ifdef LIBFAD_DO_RENDER
      cairo_line_to(render->cr, hdr->x0/20.00, hdr->y0/20.00);
#endif
      x = hdr->x0; y = hdr->y0;
      FAD_ERROR("line to (%f, %f)\n", hdr->x0/20.00, hdr->y0/20.00);
    } else {
      double x0, y0, x1, y1;
#ifdef LIBFAD_DO_RENDER
      x0 = x/20.00+(hdr->x1-x)/30.00; y0 = y/20.00+(hdr->y1-y)/30.00;
      x1 = hdr->x1/20.00+(hdr->x0-hdr->x1)/60.00; y1 = hdr->y1/20.00+(hdr->y0-hdr->y1)/60.00;
      cairo_curve_to(render->cr, x0, y0, x1, y1, hdr->x0/20.00, hdr->y0/20.00);
#endif
      x = hdr->x0; y = hdr->y0;
      FAD_ERROR("curve to (%f, %f, %f, %f)\n", hdr->x1/20.00, hdr->y1/20.00, hdr->x0/20.00, hdr->y0/20.00);
    }

    hdr = hdr->next;
  }
}

static void shape_record_render_edgesrecord(sr_edgesrecord_t *er, fad_render_t *render, shape_record_t *sr, dl_node_t *node) {
  u16_t idx = 0, count;
  u16_t total = er->fs_num+1+er->ls_num+1;

  for(; idx<total; idx++) {
  FAD_ERROR("total = %d, er = %x, array[%d] = %x\n", total, er, idx, er->array[idx]);
    if(er->array[idx]) {
      sr_node_list_t *nlptr = er->array[idx];
#ifdef LIBFAD_DEBUG
      count = 0;
#endif
      do {
        FAD_ERROR("render: node list = %x, nl->sta = %x\n", nlptr, nlptr->sta);
        shape_record_render_node_list(nlptr, render, node);

        nlptr = nlptr->next;
#ifdef  LIBFAD_DEBUG
        count++;
#endif
      } while(nlptr != er->array[idx]);

      if(node->flag&0x40)
        break;

      if(er->ns!=NULL) {
        if(idx <= er->fs_num) {
#ifdef  LIBFAD_DO_RENDER
          _fill_style_do_render(render, &(er->ns->fs_array), idx-1, sr->srv, sr->dict, node);
          cairo_fill(render->cr);
#endif
        } else {
#ifdef  LIBFAD_DO_RENDER
          _line_style_do_render(render, &(er->ns->ls_array), idx-er->fs_num-1, sr->srv, node);
          cairo_stroke(render->cr);
#endif
        }
      }
      FAD_ERROR("summary: node list count = %d\n", count);
    }
  }
}

static void shape_record_core_render(shape_record_t *sr, fad_render_t *render, dl_node_t *node) {
  sr_edgesrecord_t *ptr = sr->header;

  while(ptr != NULL) {
    if(sr->ns != NULL && ptr->ns == NULL)
      ptr->ns = sr->ns;

    FAD_ERROR("edgesrecord do render, er = %x, ns = %x\n", ptr, ptr->ns);
    shape_record_render_edgesrecord(ptr, render, sr, node);
    ptr = ptr->next;
  }
}

u8_t shape_record_do_render(shape_record_t* sr, fad_render_t* render, dl_node_t *node) {
  shape_record_core_render(sr, render, node);
  return FAD_TRUE;
}

void shape_record_free(shape_record_t* sr) {
  sr_edgesrecord_t *hdr = NULL, *ptr = NULL;

  for(hdr=sr->header; hdr!=NULL;) {
    ptr = hdr->next;
    sr_edgesrecord_destroy(hdr);
    free(hdr);
    hdr = ptr;
  }

  if(sr->ns)
    free(sr->ns);
}

static void _style_shape_do_free(fad_object_t* fo) {
  style_shape_t* ss = (style_shape_t* )fo;

  FAD_ERROR("style shape doing free, object = %x, sr = %x\n", fo, &ss->sr);
  shape_record_free(&ss->sr);
}

static u8_t _style_shape_do_render(fad_object_t* fo, fad_render_t* render, dl_node_t *node) {
  style_shape_t* ss = (style_shape_t* )fo;

  return shape_record_do_render(&ss->sr, render, node);
}

static void  _style_shape_get_rect(fad_object_t* fo, rect_t *rt) {
  style_shape_t* ss = (style_shape_t* )fo;

  *rt = ss->bound;
}
